import numpy, pandas
from sklearn.base import TransformerMixin
from sklearn.feature_selection import SelectKBest
from sklearn.feature_selection import f_regression, mutual_info_regression


class SelectNAndKBest(
	TransformerMixin,
):
	"""
	Selects the first N features plus the K best other features.
	"""

	def __init__(self, n, k, func=None):
		self._n = n
		self._k = k
		self._func = mutual_info_regression if func is None else func

	def fit(self, X, y):
		if self._k > X.shape[1]-self._n:
			use_k = 'all'
		else:
			use_k = self._k

		if X.shape[1]-self._n <=0:
			self._feature_selector = None
		elif isinstance(X, pandas.DataFrame):
			self._feature_selector = SelectKBest(self._func, k=use_k).fit(X.iloc[:,self._n:], y)
		else:
			self._feature_selector = SelectKBest(self._func, k=use_k).fit(X[:,self._n:], y)
		return self

	def transform(self, X):
		if isinstance(X, pandas.DataFrame):
			X_outside = X.iloc[:,self._n:]
			if self._feature_selector is None:
				X2 = X_outside
			else:
				X2 = self._feature_selector.transform(X_outside)
				X2 = pandas.DataFrame(X2, index=X.index, columns=X_outside.columns[self._feature_selector.get_support()])
			return pandas.concat([X.iloc[:,:self._n], X2], axis=1)
		else:
			if self._feature_selector is None:
				X2 = X[:,self._n:]
			else:
				X2 = self._feature_selector.transform(X[:,self._n:])
			return numpy.concatenate([X[:,:self._n], X2], axis=1)

